//
//  PPMTabView.m
//  PPMTabView
//
//  Created by ppm on  07/06/29.
//  Copyright ppm 2007. All rights reserved.
//
// **このサンプルコードは、Apple のサンプルコード、「Reducer (http://developer.apple.com/samplecode/Reducer/index.html)」をベースにしています。**

#import "PPMTabView.h"

// アニメーションの長さ（秒）
#define TRANSITION_DURATION 0.2

/* ======== NSAnimation のサブクラスのインターフェース ======== */
// 「NSAnimation の通常の使い方は、少なくとも setCurrentProgress:progress をオーバーライドしたサブクラスを作成する」とマニュアルにもあります。
// これを PPMTabView.m に内蔵しているのは、ベースの「Reducer」がそうしているからで、無理して内蔵する必要はありません。…というか、警告が出ますが、ここではとりあえず、放置しています。

@interface TabViewAnimation : NSAnimation

{

	int direction;
	
	#define LEFT_DIRECTION -1
	#define RIGHT_DIRECTION 1
	
}

- (int) direction;
- (void) setDirection:(int)newDirection;

@end


static void ClearBitmapImageRep(NSBitmapImageRep *bitmap) {
    unsigned char *bitmapData = [bitmap bitmapData];
    if (bitmapData != NULL) {
        bzero(bitmapData, [bitmap bytesPerRow] * [bitmap pixelsHigh]);
    }
}


/* ======== PPMTabView の実装 ======== */

@implementation PPMTabView


- (void)drawRect:(NSRect)rect {

	[super drawRect:rect]; // 先に通常通り描画する。これがないと、タブの「枠」の表示が飛んでしまう。

	if ( animation ) {
		
		NSRect clippingRect = imageRect; // タブの中身（コンテントビュー）のサイズ情報。これがないと、タブの「枠」の上から描いてしまう。
		
		// 下の selectTabViewItem:tabViewItem で用意された中間イメージを少しずつずらしながら表示するのだけど、
		// [animation currentValue] (0.0 - 1.0) でアニメーションの進行度に応じた数値が返ってくるので、ただただこれを計算式に組み込むのみ。
		if ( [animation direction] == LEFT_DIRECTION ) clippingRect.origin.x = (1 - [animation currentValue]) * (imageRect.size.width / 2);
		else clippingRect.origin.x = imageRect.size.width * [animation currentValue];
		
		clippingRect.origin.y = 0;
		
		// 描画
		[transitionImage drawInRect:imageRect fromRect:clippingRect operation:NSCompositeSourceOver fraction:1.0f];
		
    }
	
}


- (void)selectTabViewItem:(NSTabViewItem *)tabViewItem {

	NSView *initialContentView = [[self selectedTabViewItem] view]; // 現在のタブのコンテントビュー
    NSView *finalContentView = [tabViewItem view]; // 新しいタブ（選択されたタブ）のコンテントビュー
	
	NSRect rect = [initialContentView bounds];
	imageRect = [initialContentView frame];


	/* ---- 現在のタブの表示状態をビットマップイメージとして取り出す ---- */

	NSBitmapImageRep *initialContentBitmap;

	initialContentBitmap = [initialContentView bitmapImageRepForCachingDisplayInRect:rect];
	ClearBitmapImageRep(initialContentBitmap);
	[initialContentView cacheDisplayInRect:rect toBitmapImageRep:initialContentBitmap];
	
	
	// 新しいタブに切り替える前に、現在のタブ／新しいタブのインデックスを記録する。
	// これで左右どちらに動くアニメーションにするか決定する。（そういえば、３つ以上タブがある場合は想定していない。もしかしたら、このままだと違和感あるかも）
	int indexOfOldTabViewItem = [super indexOfTabViewItem:[super selectedTabViewItem]];
	int indexOfNewTabViewItem = [super indexOfTabViewItem:tabViewItem];
	
	// 新しいタブに切り替える
	[super selectTabViewItem:tabViewItem];
	
	
	/* ---- 新しいタブの表示状態をビットマップイメージとして取り出す ---- */
	
	NSBitmapImageRep *finalContentBitmap = [finalContentView bitmapImageRepForCachingDisplayInRect:rect];
    ClearBitmapImageRep(finalContentBitmap);
    [finalContentView cacheDisplayInRect:rect toBitmapImageRep:finalContentBitmap];
	
	
	/* ---- TabViewAnimation（NSAnimation のサブクラス）のインスタンスを生成 ---- */
	
	animation = [[TabViewAnimation alloc] initWithDuration:TRANSITION_DURATION animationCurve:NSAnimationEaseInOut];
    [animation setDelegate:self];
	
	
	// アニメーションの方向を指定
	if (indexOfNewTabViewItem > indexOfOldTabViewItem) [animation setDirection:RIGHT_DIRECTION];
	else [animation setDirection:LEFT_DIRECTION];
	
	
	[finalContentView setHidden:YES]; // タブの中身の表示を、いったん消す。（と言っても、まだユーザーには何の変化も見えない）
	
	
	/* ---- アニメーションに使用する画像を生成 ---- */
	// 「アニメーションに使用する画像」：具体的にどんなものかというと、先ほど取得した現在のタブ／新しいタブのビットマップイメージをアニメーションの方向に合わせて左右に並べたもの。
	
	rect.size.width += rect.size.width; // 上記の理由から、幅を２倍に。流れで変数 rect を再利用しているけど、若干行儀が良くない気もする（汗）
	
	transitionImage = [[NSImage alloc] initWithSize:rect.size]; // 「アニメーションの方向」はアニメーションオブジェクトが持っておくべき、と思ったので TabViewAnimation を拡張したのだが、どちらかというと、これもか。
	[transitionImage setFlipped:YES];
	
	// 描画
	[transitionImage lockFocus];
	
	if ( [animation direction] == LEFT_DIRECTION ) [transitionImage drawRepresentation:finalContentBitmap inRect:rect];
	else [transitionImage drawRepresentation:initialContentBitmap inRect:rect];
	
	rect.origin.x += rect.size.width / 2;
	
	if ( [animation direction] == LEFT_DIRECTION ) [transitionImage drawRepresentation:initialContentBitmap inRect:rect];
	else [transitionImage drawRepresentation:finalContentBitmap inRect:rect];
	
	[transitionImage unlockFocus];
	
	
	/* ---- アニメーション開始 ---- */
	
	[animation startAnimation];
	
	
	/* ---- 後片付け ---- */
	// アニメーションが終了すると、ここに戻ってくる。
	// こういう方法の他に、アニメーションを開始させたままほうっておいて、animation のデリゲートで animationDidEnd: を受け取る、という方法もある。
	
	[animation release];
    animation = nil;
	[transitionImage release];
	transitionImage = nil;
    [finalContentView setHidden:NO];
    [self setNeedsDisplay:YES];

}


@end


/* ======== NSAnimation のサブクラスの実装 ======== */
// サンプル「Reducer」で初めて知ったのですが、こうやって別のクラスの定義を内蔵できる（！？）
// でも、「Reducer」のようにインスタンス変数も何も無い状態なら良いようですが、このように拡張していくと、どうもコンパイラに警告されてしまいます。

@implementation TabViewAnimation

- (void)setCurrentProgress:(NSAnimationProgress)progress {

    [super setCurrentProgress:progress];

	[[self delegate] display];
	
}

- (int) direction {
	return direction;
}

- (void) setDirection:(int)newDirection {
	direction = newDirection;
}

@end